# Copyright (C) 2025 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import sys
import requests
from typing import Any
from requests.exceptions import HTTPError
from requests.exceptions import RequestException
from urllib.parse import urlencode
from datetime import datetime
from sdx_dl.sdxparser import logger, args
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl

ua = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:125.0) Gecko/20100101 Firefox/125.0"

if args.proxy:
    proxie = f"{args.proxy}"
    if not (any(p in proxie for p in ["http", "https"])):
        proxie = f"http://{proxie}"
    proxies = {'http': proxie, 'https': proxie}
else:
    proxies = None

__all__ = ['SubxAPI']


def ExceptionErrorMessage(e: Exception):
    """Parse ``Exception`` error message."""
    if isinstance(e, (HTTPError, RequestException)):
        msg = e.__str__().split(":", maxsplit=1)[1].split("(")[0].strip()
    else:
        msg = e
    error_class = e.__class__.__name__
    console.print(f':no_entry: {gl("Error_occurred")} {gl(error_class)} : {msg}')


class SubxAPI:
    """Base API for SubX"""
    def __init__(self, token: str, default_timeout: int = 15):
        """
        Initialize the API

        Args:
            token: The Bearer token for authentication
            default_timeout: Default timeout in seconds for requests
            base_url: The base URL of the API
        """
        self.base_url = "https://subx-api.duckdns.org/api"
        self.token = token
        self.default_timeout = default_timeout
        self.session = requests.Session()
        self.session.headers.update({
            'Authorization': f'Bearer {self.token}',
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'User-Agent': ua
        })
        # Setting proxy
        if proxies:
            self.session.proxies.update(proxies)
        # Data storage of query request
        self._data: dict[str, Any] | None = None

    def query(
        self,
        query: str = "",
        method: str = "GET",
        endpoint: str = "subtitles/search/",
        params: dict[str, Any] | None = None,
        timeout: int | None = None
    ) -> dict[str, Any] | str | None:
        """
        Make a query request to the API

        Args:
            method: HTTP method (GET)
            endpoint: API endpoint (appended to base_url)
            params: Query parameters
            timeout: Request timeout in seconds

        Returns:
            Parsed JSON response if content is JSON, otherwise raw text,
            or None if request fails
        """
        url = f"{self.base_url}/{endpoint.rstrip('/')}"
        timeout = timeout or self.default_timeout

        response: requests.Response
        try:
            response = self.session.request(
                method,
                url,
                params=params or urlencode({"query": query}),
                timeout=timeout
            )
            response.raise_for_status()

            # Try to parse JSON, fallback to text
            try:
                self._data = response.json()
                return self._data
            except ValueError:
                return response.text

        except HTTPError as e:
            ExceptionErrorMessage(e)
            logger.debug(f"HTTP error occurred: {e} HTTP Error ({e.response.status_code})")
            sys.exit(1)
        except RequestException as err:
            ExceptionErrorMessage(err)
            logger.debug(f"Request error occurred: {err}")
            sys.exit(1)
        except Exception as err:
            console.print(
                f':no_entry: {gl("Unexpected_error")}: {err.__str__()}',
                emoji=True, new_line_start=True
            )
            sys.exit(1)

        return None

    def get(
        self,
        id: str = "",
        method: str = "GET",
        endpoint: str = "subtitles/",
        timeout: int | None = None
    ) -> requests.Response | Any | None:
        """
        Make an request to the API

        Args:
            method: HTTP method (GET)
            endpoint: API endpoint (appended to base_url)
            id: subtitle id parameter
            timeout: Request timeout in seconds

        Returns:
            Parsed JSON response if content is JSON, otherwise raw text,
            or None if request fails
        """
        url = f"{self.base_url}/{endpoint}/{id}/download"
        timeout = timeout or self.default_timeout
        self.session.headers.update({
            'Accept': '*/*',
            'Content-Type': 'application/x-www-form-urlencoded',
            'content-disposition': 'attachment'
        })

        response: requests.Response
        try:
            response = self.session.request(
                method,
                url,
                timeout=timeout
            )
            response.raise_for_status()

            # Try to get content response or None
            try:
                return response
            except Exception:
                return None

        except HTTPError as e:
            ExceptionErrorMessage(e)
            logger.debug(f"HTTP error occurred: {e} HTTP Error ({e.response.status_code})")
            sys.exit(1)
        except RequestException as err:
            ExceptionErrorMessage(err)
            logger.debug(f"Request error occurred: {err}")
            sys.exit(1)
        except Exception as err:
            console.print(
                f':no_entry: {gl("Unexpected_error")}: {err.__str__()}',
                emoji=True, new_line_start=True
            )
            logger.debug(f'{gl("Unexpected_error")}: {err}')
            sys.exit(1)

        return None

    def from_subx_aadata(self) -> dict[str, Any]:
        """convert subx data to aadata format"""
        data: dict[str, Any] = {}
        list_data: list[dict[str, Any]] = []

        if self._data and bool(self._data):
            total = int(self._data['total'])
            for item in self._data['items']:
                dt_object = datetime.strptime(item['posted_at'], "%Y-%m-%dT%H:%M:%SZ")
                posted_at = dt_object.strftime("%Y-%m-%d %H:%M:%S")
                data = {
                    "id": item['id'],
                    "titulo": item['title'],
                    "descripcion": item['description'],
                    "descargas": item['downloads'],
                    "nick": item['uploader_name'],
                    "fecha_subida": posted_at,
                    "comentarios": 0
                }
                list_data.append(data)

            return {
                "iTotalRecords": total,
                "aaData": list_data
            }
        else:
            return {}



# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import tempfile
import argparse
import logging
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import set_locale, gl
from sdx_dl.sdxclasses import (
    ChkVersionAction,
    ConfigManager,
    ViewConfigAction,
    SaveConfigAction,
    SetConfigAction,
    ResetConfigAction,
    BypasserAction,
    SetBypasserConfigAction,
    validate_proxy
)
from importlib.metadata import version
from rich.logging import RichHandler
from rich.traceback import install
install(show_locals=True)

__all__ = ["args", "logger"]


def create_parser():
    parser = argparse.ArgumentParser(
        prog='sdx-dl',
        formatter_class=argparse.RawTextHelpFormatter,
        usage="sdx-dl [options] search",
        description='A cli tool for download subtitle from https://www.subdivx.com '
        'with the better possible matching results.',
        epilog='Project issues:https://github.com/Spheres-cu/subdx-dl/issues\n'
        '\nUsage examples:https://github.com/Spheres-cu/subdx-dl#examples\n\n'
        'Wiki:https://github.com/Spheres-cu/subdx-dl/wiki\n\n'
    )
    parser.add_argument('search', type=str, help="file, directory or movie/series"
                        " title or IMDB Id to retrieve subtitles")
    parser.add_argument('--quiet', '-q', action='store_true', default=False, help="No verbose mode")
    parser.add_argument('--verbose', '-v', action='store_true', default=False, help="Be in verbose mode")
    parser.add_argument('--force', '-f', action='store_true', default=False, help="override existing file")
    parser.add_argument('--no-choose', '-nc', action='store_true', default=False, help="No Choose sub manually")
    parser.add_argument('--no-filter', '-nf', action='store_true', default=False, help="Do not filter search results")
    parser.add_argument('--nlines', '-nl', type=int, choices=[5, 10, 15, 20], default=False, nargs='?', const=10,
                        help="Show nl(5,10,15,20) availables records per screen. Default 10 records.", metavar="")
    parser.add_argument('--lang', '-l', type=str, choices=["es", "en"], default=False, nargs='?', const="es",
                        help="Show messages in language es or en", metavar="")
    parser.add_argument('--version', '-V', action='version',
                        version=f'subdx-dl {version("subdx-dl")}', help="Show program version")
    parser.add_argument('--check-version', '-cv', action=ChkVersionAction, help="Check for new version")

    # Download opts group
    download_opts = parser.add_argument_group('Download')
    download_opts.add_argument('--path', '-p', type=str, help="Path to download subtitles")
    download_opts.add_argument('--proxy', '-x', type=str, help="Set a http(s) proxy(x) connection", metavar="x")

    # Search opts group
    search_opts = parser.add_argument_group('Search by')
    search_opts.add_argument('--Season', '-S', action='store_true', default=False, help="Search by Season")
    search_opts.add_argument('--kword', '-k', type=str, help="Add keywords to search among subtitles descriptions", metavar="kw")
    search_opts.add_argument('--title', '-t', type=str, help="Set the title to search", metavar="t")
    search_opts.add_argument('--imdb', '-i', action='store_true', default=False, help="Search first for the IMDB id or title")
    search_opts.add_argument('--SubX', '-sx', action='store_true', default=False, help="Search using SubX API")

    # Config opts group
    config_opts = parser.add_argument_group('Config').add_mutually_exclusive_group()
    config_opts.add_argument('--view-config', '-vc', action=ViewConfigAction, help="View config file")
    config_opts.add_argument('--save-config', '-sc', action=SaveConfigAction, help="Save options to config file")
    config_opts.add_argument('--load-config', '-lc', action='store_true', default=False, help="Load config file options")

    config_opts.add_argument(
        '--config', '-c', action=SetConfigAction,
        choices=["quiet", "verbose", "force", "no_choose", "no_filter", "nlines", "path", "proxy", "Season", "imdb", "lang", "SubX"],
        nargs='?', metavar="o", help="Save an option[o] to config file"
    )
    config_opts.add_argument(
        '--reset', '-r', action=ResetConfigAction,
        choices=["quiet", "verbose", "force", "no_choose", "no_filter", "nlines", "path", "proxy", "Season", "imdb", "lang", "SubX"],
        metavar="o", help="Reset an option[o] in the config file"
    )

    # Bypasser opts
    bypasser_opts = parser.add_argument_group('Bypasser').add_mutually_exclusive_group()
    bypasser_opts.add_argument(
        '--bypass', '-b',
        action=BypasserAction,
        choices=["force", "manual"], default="",
        nargs='?', metavar="o",
        help="Run bypass with options [force, manual]"
    )
    bypasser_opts.add_argument(
        '--conf-bypass', '-cb',
        action=SetBypasserConfigAction,
        help='Config bypass options'
    )

    return parser


def create_logger(level: str = "DEBUG", verbose: bool = False):

    # Setting logger
    levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
    LOGGER_LEVEL = levels[4]
    LOGGER_FORMATTER_LONG = logging.Formatter('%(asctime)-12s %(levelname)-6s %(message)s', '%Y-%m-%d %H:%M:%S')
    LOGGER_FORMATTER_SHORT = logging.Formatter(fmt='%(message)s', datefmt="[%X]")

    level = level if level in levels else LOGGER_LEVEL
    temp_log_dir = tempfile.gettempdir()
    file_log = os.path.join(temp_log_dir, 'subdx-dl.log')

    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(level)

    if not verbose:
        logfile = logging.FileHandler(file_log, mode='w', encoding='utf-8')
        logfile.setFormatter(LOGGER_FORMATTER_LONG)
        logfile.setLevel(level)
        logger.addHandler(logfile)
    else:
        console = RichHandler(rich_tracebacks=True, tracebacks_show_locals=True)
        console.setFormatter(LOGGER_FORMATTER_SHORT)
        console.setLevel(level)
        logger.addHandler(console)


parser = create_parser()
args = parser.parse_args()
cf = ConfigManager()

if args.load_config:
    if cf.exists and cf.hasconfig:
        copied_args = args.__dict__.copy()
        new_args = cf.merge_config(copied_args)

        for k, v in new_args.items():
            args.__setattr__(k, v)

if args.verbose:
    args.__setattr__("quiet", True)
create_logger(verbose=args.verbose)

if cf.hasconfig and 'SubX' in cf.config:
    args.__setattr__("SubX", cf.get("SubX"))

if args.load_config:
    logger.debug("Config loaded!")

if args.lang:
    set_locale(args.lang)
else:
    if cf.hasconfig and 'lang' in cf.config:
        set_locale(cf.get("lang", "es"))

if args.path:
    if not (os.path.isdir(args.path) and os.access(args.path, os.W_OK)):
        if args.quiet:
            logger.debug(f'Directory {args.path} do not exists')
        else:
            console.print(
                f':no_entry: [bold red]{gl("Directory")}[/]{args.path} [bold red]{gl("Directory_not_exists")}[/]',
                new_line_start=True, emoji=True
            )
        sys.exit(1)

if (args.proxy and not validate_proxy(args.proxy)):
    if args.quiet:
        logger.debug("Incorrect proxy setting. Only http, https or IP/domain:PORT is accepted")
    else:
        console.print(
            f':no_entry: [bold red]{gl("Incorrect_proxy_setting").split(".")[0]} :[yellow]{args.proxy}[/]',
            new_line_start=True, emoji=True
        )
    sys.exit(1)
else:
    if cf.hasconfig and 'proxy' in cf.config:
        args.__setattr__("proxy", cf.get("proxy"))



import random

OPERATING_SYSTEMS = ["windows", "macos", "linux"]

SCREEN_RESOLUTIONS = [
    (1920, 1080), (1920, 1200), (2560, 1440), (1680, 1050),
    (1600, 900), (1366, 768), (1440, 900), (1536, 864),
    (2560, 1600), (3840, 2160)
]


class BrowserConfig:
    """Generate random browser configurations with integrated UA generation."""
        
    @staticmethod
    def get_firefox_headers() -> dict[str, str]:
        """Get Firefox-specific headers."""
        return {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8',
            'Accept-Language': 'en-US,en;q=0.5',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
            'Upgrade-Insecure-Requests': '1',
            'Cache-Control': 'max-age=0',
        }
    
    @staticmethod
    def generate_random_config(
        selected_os: str | None = None, 
        firefox_version: int | None = None, 
        lang: str = "en-US"
    ) -> dict:
        """Generate random browser configuration based on OS."""
        
        if selected_os is None or selected_os not in OPERATING_SYSTEMS:
            selected_os = random.choice(OPERATING_SYSTEMS)
        
        if firefox_version is None:
            firefox_version = random.randint(140, 145) # Recent Firefox versions
        
        # Random screen resolution
        screen_width, screen_height = random.choice(SCREEN_RESOLUTIONS)
        
        # Calculate inner dimensions (subtract browser chrome)
        toolbar_height = random.randint(70, 100)
        inner_width = screen_width
        inner_height = screen_height - toolbar_height
        
        # Random history length
        history_length = random.randint(2, 8)
        
        # Random hardware cores (realistic values)
        hardware_cores = random.choice([2, 4, 6, 8, 12, 16, 20, 24])
        
        # Language configuration
        languages = [lang]
        if lang != "en-US" and not lang.startswith("en"):
            languages.append("en-US")
        
        config = {
            'window.outerHeight': screen_height,
            'window.outerWidth': screen_width,
            'window.innerHeight': inner_height,
            'window.innerWidth': inner_width,
            'window.history.length': history_length,
            'navigator.appCodeName': 'Mozilla',
            'navigator.appName': 'Netscape',
            'navigator.hardwareConcurrency': hardware_cores,
            'navigator.product': 'Gecko',
            'navigator.productSub': '20100101',
            'navigator.language': lang,
            'navigator.languages': languages,
        }
        
        # OS-specific configurations
        if selected_os == "windows":
            return BrowserConfig._configure_windows(config, firefox_version)
        elif selected_os == "macos":
            return BrowserConfig._configure_macos(config, firefox_version)
        elif selected_os == "linux":
            return BrowserConfig._configure_linux(config, firefox_version)
        
        return config
    
    @staticmethod
    def _configure_windows(config: dict, firefox_version: int) -> dict:
        """Configure Windows-specific browser settings."""
        win_versions = [
            "Windows NT 10.0; Win64; x64",
            "Windows NT 11.0; Win64; x64",
        ]
        win_version = random.choice(win_versions)
        
        config.update({
            'navigator.userAgent': f'Mozilla/5.0 ({win_version}; rv:{firefox_version}.0) Gecko/20100101 Firefox/{firefox_version}.0',
            'navigator.appVersion': f'5.0 ({win_version})',
            'navigator.oscpu': win_version,
            'navigator.platform': 'Win32',
            'navigator.maxTouchPoints': random.choice([0, 10]),
        })
        return config
    
    @staticmethod
    def _configure_macos(config: dict, firefox_version: int) -> dict:
        """Configure macOS-specific browser settings."""
        macos_versions = [
            ("13_0", "13.0"),  # Ventura
            ("14_0", "14.0"),  # Sonoma
            ("15_0", "15.0"),  # Sequoia
        ]
        mac_ver_underscore, mac_ver_dot = random.choice(macos_versions)
        
        config.update({
            'navigator.userAgent': f'Mozilla/5.0 (Macintosh; Intel Mac OS X {mac_ver_underscore}; rv:{firefox_version}.0) Gecko/20100101 Firefox/{firefox_version}.0',
            'navigator.appVersion': f'5.0 (Macintosh)',
            'navigator.oscpu': f'Intel Mac OS X {mac_ver_dot}',
            'navigator.platform': 'MacIntel',
            'navigator.maxTouchPoints': 0,
        })
        return config
    
    @staticmethod
    def _configure_linux(config: dict, firefox_version: int) -> dict:
        """Configure Linux-specific browser settings."""
        linux_distros = [
            "X11; Linux x86_64",
            "X11; Ubuntu; Linux x86_64", 
            "X11; Fedora; Linux x86_64",
            "X11; Debian; Linux x86_64",
            "X11; CentOS; Linux x86_64",
            "X11; Arch Linux; Linux x86_64",
            "X11; openSUSE; Linux x86_64",
            "X11; Manjaro; Linux x86_64",
        ]
        linux_distro = random.choice(linux_distros)
        
        config.update({
            'navigator.userAgent': f'Mozilla/5.0 ({linux_distro}; rv:{firefox_version}.0) Gecko/20100101 Firefox/{firefox_version}.0',
            'navigator.appVersion': '5.0 (X11)',
            'navigator.oscpu': 'Linux x86_64',
            'navigator.platform': 'Linux x86_64',
            'navigator.maxTouchPoints': 0,
        })
        return config


# Maintain backward compatibility
def generate_random_config(selected_os: str | None = None, firefox_version: int | None = None, lang: str = "en-US") -> dict:
    """Generate random browser configuration based on OS (backward compatibility function)."""
    return BrowserConfig.generate_random_config(selected_os, firefox_version, lang)


import json
import os
import threading

from sdx_dl.cf_bypasser.utils.misc import logger
from datetime import datetime, timedelta
from dataclasses import dataclass
from typing import Any


@dataclass
class CachedCookies:
    """Represents cached Cloudflare clearance data."""
    hostname: str
    cookies: dict[str, str]
    user_agent: str
    timestamp: datetime
    expires_at: datetime

    def is_expired(self) -> bool:
        """Check if the cached cookies are expired."""
        return datetime.now() >= self.expires_at

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary for serialization."""
        return {
            'hostname': self.hostname,
            'cookies': self.cookies,
            'user_agent': self.user_agent,
            'timestamp': self.timestamp.isoformat(),
            'expires_at': self.expires_at.isoformat()
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> 'CachedCookies':
        """Create from dictionary for deserialization."""
        return cls(
            hostname=data['hostname'],
            cookies=data['cookies'],
            user_agent=data['user_agent'],
            timestamp=datetime.fromisoformat(data['timestamp']),
            expires_at=datetime.fromisoformat(data['expires_at'])
        )


class CookieCache:
    """Thread-safe cache for Cloudflare clearance cookies."""

    def __init__(self, cache_file: str = "cf_cookie_cache.json"):
        self.cache_file = cache_file
        self.cache: dict[str, CachedCookies] = {}
        self.lock = threading.RLock()
        self._load_cache()

    def _load_cache(self):
        """Load cache from file."""
        try:
            if os.path.exists(self.cache_file):
                with open(self.cache_file, 'r') as f:
                    data = json.load(f)
                    for hostname, cached_data in data.items():
                        try:
                            self.cache[hostname] = CachedCookies.from_dict(cached_data)
                        except Exception as e:
                            logger.warning(f"Failed to load cached data for {hostname}: {e}")
        except Exception as e:
            logger.warning(f"Failed to load cache file: {e}")

    def _save_cache(self):
        """Save cache to file."""
        try:
            with open(self.cache_file, 'w') as f:
                data = {hostname: cached.to_dict() for hostname, cached in self.cache.items()}
                json.dump(data, f, indent=2)
        except Exception as e:
            logger.error(f"Failed to save cache file: {e}")

    def get(self, hostname: str) -> CachedCookies | None:
        """Get cached cookies for hostname."""
        with self.lock:
            cached = self.cache.get(hostname)
            if cached and not cached.is_expired():
                logger.info(f"Using cached cookies for {hostname}")
                return cached
            elif cached and cached.is_expired():
                logger.info(f"Cached cookies for {hostname} expired, removing")
                del self.cache[hostname]
                self._save_cache()
            return None

    def set(self, hostname: str, cookies: dict[str, str], user_agent: str, ttl_hours: int = 2):
        """Cache cookies for hostname with TTL."""
        with self.lock:
            expires_at = datetime.now() + timedelta(hours=ttl_hours)
            cached = CachedCookies(
                hostname=hostname,
                cookies=cookies,
                user_agent=user_agent,
                timestamp=datetime.now(),
                expires_at=expires_at
            )
            self.cache[hostname] = cached
            self._save_cache()
            logger.info(f"Cached cookies for {hostname}, expires at {expires_at}")

    def clear_expired(self):
        """Remove all expired entries."""
        with self.lock:
            expired_keys = [k for k, v in self.cache.items() if v.is_expired()]
            for key in expired_keys:
                del self.cache[key]
            if expired_keys:
                self._save_cache()
                logger.info(f"Cleared {len(expired_keys)} expired cache entries")

    def invalidate(self, hostname: str):
        """Invalidate cached cookies for a specific hostname."""
        with self.lock:
            if hostname in self.cache:
                del self.cache[hostname]
                self._save_cache()
                logger.info(f"Invalidated cache for {hostname}")

    def clear_all(self):
        """Clear all cached entries."""
        with self.lock:
            self.cache.clear()
            self._save_cache()
            logger.info("Cleared all cache entries")



# type:ignore
import time
from DrissionPage import ChromiumPage
from sdx_dl.cf_bypasser.utils.misc import logger


class CloudflareBypasser:
    def __init__(self, driver: ChromiumPage, max_retries=-1, log=True):
        self.driver = driver
        self.max_retries = max_retries
        self.log = log

    def search_recursively_shadow_root_with_iframe(self, ele):
        if ele.shadow_root:
            if ele.shadow_root.child().tag == "iframe":
                return ele.shadow_root.child()
        else:
            for child in ele.children():
                result = self.search_recursively_shadow_root_with_iframe(child)
                if result:
                    return result
        return None

    def search_recursively_shadow_root_with_cf_input(self, ele):
        if ele.shadow_root:
            if ele.shadow_root.ele("tag:input"):
                return ele.shadow_root.ele("tag:input")
        else:
            for child in ele.children():
                result = self.search_recursively_shadow_root_with_cf_input(child)
                if result:
                    return result
        return None

    def locate_cf_button(self):
        button = None
        eles = self.driver.eles("tag:input")
        for ele in eles:  # type: ignore
            if "name" in ele.attrs.keys() and "type" in ele.attrs.keys():
                if "turnstile" in ele.attrs["name"] and ele.attrs["type"] == "hidden":
                    button = ele.parent().shadow_root.child()("tag:body").shadow_root("tag:input")
                    break

        if button:
            return button
        else:
            # If the button is not found, search it recursively
            self.log_message("Basic search failed. Searching for button recursively.")
            ele = self.driver.ele("tag:body")
            iframe = self.search_recursively_shadow_root_with_iframe(ele)
            if iframe:
                button = self.search_recursively_shadow_root_with_cf_input(iframe("tag:body"))
            else:
                self.log_message("Iframe not found. Button search failed.")
            return button

    def log_message(self, message):
        if self.log:
            logger.debug(message)

    def click_verification_button(self):
        try:
            button = self.locate_cf_button()
            if button:
                self.log_message("Verification button found. Attempting to click.")
                button.click()
            else:
                self.log_message("Verification button not found.")

        except Exception as e:
            self.log_message(f"Error clicking verification button: {e}")

    def is_bypassed(self):
        try:
            title = self.driver.title.lower()
            return "just a moment" not in title
        except Exception as e:
            self.log_message(f"Error checking page title: {e}")
            return False

    def bypass(self):

        try_count = 0

        while not self.is_bypassed():
            if 0 < self.max_retries + 1 <= try_count:
                self.log_message("Exceeded maximum retries. Bypass failed.")
                break

            self.log_message(f"Attempt {try_count + 1}: Verification page detected. Trying to bypass...")
            self.click_verification_button()

            try_count += 1
            time.sleep(2)

        if self.is_bypassed():
            self.log_message("Bypass successful.")
        else:
            self.log_message("Bypass failed.")



# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import re
import sys
import json
import random
import argparse
import urllib3
import certifi
import typing

from pathlib import Path
from typing import Any
from guessit import guessit, jsonutils  # type: ignore
from bs4 import BeautifulSoup, Tag
from importlib.metadata import version
from urllib3.exceptions import HTTPError
from pygments.lexers import guess_lexer
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl, set_locale


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

__all__ = [
    "HTML2BBCode",
    "NoResultsError",
    "GenerateUserAgent",
    "VideoMetadataExtractor",
    "validate_proxy",
    "ChkVersionAction",
    "ConfigManager",
    "ViewConfigAction",
    "SaveConfigAction",
    "SetConfigAction",
    "ResetConfigAction",
    "FindFiles"
]


#  HTML2BBCode class
class HTML2BBCode:
    """
    HTML to BBCode converter
    """
    @staticmethod
    def guess_language(code: str) -> str:
        """
        Guesses the programming language from a string. This is not accurate for short strings. 'pygments' is used to
        guess code here because many other language detectors cannot detect the language properly.
        """
        guessed = guess_lexer(code).name
        if guessed == "Text only":
            return ""
        return guessed.lower()

    @classmethod
    def _html_to_bbcode(cls, tag: Tag) -> str:
        result = ""
        if hasattr(tag, 'children'):
            for child in tag.children:
                if isinstance(child, str):
                    result += child
                elif isinstance(child, Tag):
                    result += cls._html_to_bbcode(child)

        if tag.name == "span":
            if "class" in list(tag.attrs.keys()):
                tag_type = tag["class"][0]
                if tag_type == "bb-bold":  # bold
                    result = f"[bold green]{result}[/bold green]"
                elif tag_type == "bb-italic":  # italic
                    result = f"[italic bright_yellow]{result}[/italic bright_yellow]"
                elif tag_type == "bb-underline":  # underline
                    result = f"[u]{result}[/u]"
                elif tag_type == "bb-strikethrough":  # strikethrough
                    result = f"[s]{result}[/s]"
                elif tag_type == "bb-big":  # big
                    result = f"[big]{result}[/big]"
                elif tag_type == "bb-small":  # small
                    result = f"[small]{result}[/small]"
            elif "style" in list(tag.attrs.keys()):
                if "color" in tag["style"]:  # color
                    tag_color = f'{tag.get("style")}'
                    result = f'[color={tag_color.replace("color:", "")}]{result}[/color]'
        elif tag.name == "br":  # line break
            result = "\n"
        elif tag.name == "a":  # url
            result = f"[link={tag['href']}]{result}[/link]"
        elif tag.name == "b":  # bold
            result = f"[bold green]{result}[/bold green]"
        elif tag.name == "i":  # bold
            result = f"[italic bright_yellow]{result}[/italic bright_yellow]"  # italic
        elif tag.name == "ul":  # unordered list
            result = f"[list]\n{result}[/list]"
        elif tag.name == "li":  # list item
            result = f"[*] {result}"
        elif tag.name == "ol":  # ordered list
            result = f"[list=1]\n{result}[/list]"
        elif tag.name == "blockquote":  # quote
            quote_author = tag.find("p", {"class": "bb-quote-author"})
            if quote_author is not None:
                quote_author_name = str(quote_author.get_text()).replace(" wrote:", "")
                quote_author = "=" + quote_author_name
            else:
                quote_author = ""
            result = f"[quote{quote_author}]{result}[/quote]"
        elif tag.name == "p":
            if "class" in list(tag.attrs.keys()):
                if "bb-quote-author" in tag["class"]:
                    result = ""
            else:
                result = f"[p]{result}[/p]"  # p
        elif tag.name == "div":
            if "class" in list(tag.attrs.keys()):
                if "code" in tag["class"]:  # code
                    language = cls.guess_language(tag.get_text())
                    if not language:
                        language = ""
                    else:
                        language = "=" + language
                    result = f"[code{language}]{result}[/code]"
                # This converter cannot convert scratchblocks html to scratchblocks bbcode
                elif "scratchblocks" in tag["class"]:  # scratchblocks
                    result = f"[scratchblocks]{tag.get_text()}[/scratchblocks]"
            elif "style" in list(tag.attrs.keys()):
                if "text-align:center;" in tag["style"]:  # center
                    result = f"[center]{result}[/center]"
        return result

    @classmethod
    def html_to_bbcode(cls, html: str) -> str:
        try:
            soup = BeautifulSoup(html, "lxml")  # lxml is the fastest
            return cls._html_to_bbcode(soup)
        except Exception:
            pass
            return html


#  Utils Classes
class NoResultsError(Exception):
    pass


# Generate a user agent class
class GenerateUserAgent:
    """
    Class containing methods for generating user agents.
    """

    @staticmethod
    def _token() -> str:
        return "Mozilla/5.0"

    @staticmethod
    def _platform() -> str:
        _WINDOWS_PREFIX: str = "Windows NT 10.0; Win64; x64"
        _MAC_PREFIX: str = "Macintosh; Intel Mac OS X"
        _LINUX_PREFIX: str = "X11; Ubuntu; Linux x86_64"

        if sys.platform == "win32":
            # Windows
            platform = _WINDOWS_PREFIX
        elif sys.platform == "darwin":
            # macOS
            platform = _MAC_PREFIX
        else:
            # Linux and other UNIX-like systems
            platform = _LINUX_PREFIX
        return f'{platform}'

    @classmethod
    def firefox(cls) -> list[str]:
        """Generate a list of common firefox user agents

        Returns:
            list[str]: The list of common firefox user agents
        """
        return [
            f"{cls._token()} ({cls._platform()}; rv:{version}.0) Gecko/20100101 Firefox/{version}.0"
            for version in range(120, 138)
        ]

    @classmethod
    def chrome(cls) -> list[str]:
        """Generate a list of common chrome user agents

        Returns:
            list[str]: The list of common chrome user agents
        """
        return [
            f"{cls._token()} ({cls._platform()}) AppleWebKit/537.36 (KHTML, like Gecko) \
            Chrome/{version}.0.0.0 Safari/537.36"
            for version in range(120, 135)
        ]

    @classmethod
    def opera(cls) -> list[str]:
        """Generate a list of common opera user agents

        Returns:
            list[str]: The list of common opera user agents
        """
        return [
            f"{cls._token()} ({cls._platform()}) AppleWebKit/537.36 (KHTML, like Gecko) \
            Chrome/{version}.0.0.0 Safari/537.36 OPR/{opr}.0.0.0"
            for version in range(120, 135, 5) for opr in range(103, 118, 5)
        ]

    @classmethod
    def safari(cls) -> list[str]:
        """Generate a list of common safari user agents

        Returns:
            list[str]: The list of common safari user agents
        """
        if sys.platform == "darwin":
            return [
                f"{cls._token()} ({cls._platform()} 14_7_5) AppleWebKit/605.1.15 (KHTML, like Gecko) \
                Version/{major}.{minor} Safari/605.1.15"
                for major, minors in [(16, range(5, 7)), (17, range(0, 7))] for minor in minors
            ]
        else:
            return []

    @classmethod
    def safari_mobile(cls) -> list[str]:
        """Generate a list of common mobile safari user agents

        Returns:
            list[str]: The list of common safari mobile user agents
        """
        if sys.platform == "darwin":
            return [
                f"{cls._token()} (iPhone; CPU iPhone OS {major}_{minor} like Mac OS X) \
                AppleWebKit/605.1.15 (KHTML, like Gecko) Version/{major}.{minor} Mobile/15E148 Safari/604.1"
                for major, minors in [(16, range(5, 8)), (17, range(0, 7))] for minor in minors
            ]
        else:
            return []

    @staticmethod
    def generate_all() -> list[str]:
        """Convenience method, Generate common user agents for all supported browsers

        Returns:
            list[str]: The list of common user agents for all supported browsers in GenerateUserAgent.
        """
        if sys.platform == "darwin":
            return GenerateUserAgent.safari() + GenerateUserAgent.safari_mobile() + GenerateUserAgent.opera()
        else:
            return GenerateUserAgent.firefox() + GenerateUserAgent.chrome() + GenerateUserAgent.opera()

    @staticmethod
    def generate_random() -> list[str]:
        """Convenience method, Generate random user agents for all supported browsers

        Returns:
            list[str]: The list of random user agents for all supported browsers in GenerateUserAgent.
        """
        if sys.platform == "darwin":
            return random.choice([GenerateUserAgent.safari() + GenerateUserAgent.safari_mobile() + GenerateUserAgent.opera()])
        else:
            return random.choice([GenerateUserAgent.firefox(), GenerateUserAgent.chrome(), GenerateUserAgent.opera()])

    @staticmethod
    def random_browser() -> str:
        """Convenience method, Generate a random user agents for one supported browser

        Returns:
            str: With the random user agents for one supported browser in GenerateUserAgent.
        """
        if sys.platform == "darwin":
            browser = random.choice([GenerateUserAgent.safari() + GenerateUserAgent.safari_mobile() + GenerateUserAgent.opera()])
            return random.choice(browser)
        else:
            browser = random.choice([GenerateUserAgent.firefox(), GenerateUserAgent.chrome(), GenerateUserAgent.opera()])
            return random.choice(browser)


# validate proxy settings
def validate_proxy(proxy_str: str) -> bool:
    """
    Validation with IP address or domain and port.
    """

    ip_pattern = r'^(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'
    host_pattern = r'^[a-zA-Z0-9][a-zA-Z0-9.-]*[a-zA-Z0-9]$'

    match = re.match(r'^(?:(https|http)://)?(?:([^:@]+):([^:@]+)@)?([^:@/]+)(?::(\d+))?$', proxy_str)

    if not match:
        return False

    protocol, _, _, host, port = match.groups()

    if not (re.match(ip_pattern, host) or re.match(host_pattern, host)):
        return False

    if (port is None) or (not (0 < int(port) <= 65535)):
        return False

    if protocol not in ["http", "https", None]:
        return False

    return True


# Check version
def ExceptionErrorMessage(e: Exception):
    """Parse ``Exception`` error message."""
    if isinstance(e, (HTTPError)):
        msg = e.__str__().split(":", maxsplit=1)[1].split("(")[0]
    else:
        msg = e.__str__()
    error_class = e.__class__.__name__
    console.print(gl("Error_occurred") + error_class + ":" + msg)
    sys.exit(1)


def _check_version(version: str, proxy: str):
    """Check for new version."""
    ua = GenerateUserAgent.random_browser()
    headers = {"user-agent": ua}
    PYPI_API_URL = 'https://pypi.org/pypi/subdx-dl/json'
    GITHUB_API_URL = 'https://api.github.com/repos/spheres-cu/subdx-dl/releases/latest'
    msg = ""

    if (proxy):
        if not (any(p in proxy for p in ["http", "https"])):
            proxy = "http://" + proxy
        session = urllib3.ProxyManager(
            proxy, headers=headers, cert_reqs="CERT_REQUIRED",
            ca_certs=certifi.where(), retries=2, timeout=10
        )
    else:
        session = urllib3.PoolManager(
            headers=headers, cert_reqs="CERT_REQUIRED",
            ca_certs=certifi.where(), retries=2, timeout=10
        )

    def get_version_description():
        """Get new `version` description."""
        data: dict[str, Any] = {}
        try:
            response = session.request('GET', GITHUB_API_URL).data
            data = json.loads(response)
        except (HTTPError, Exception) as e:
            ExceptionErrorMessage(e)
        if bool(data):
            description = f'{data.get("body", "")}'
        else:
            description = ""

        return description

    try:
        _dt_version = session.request('GET', PYPI_API_URL).data
        data: dict[str, Any] = json.loads(_dt_version)['info']
        _pypi_version = f'{data.get("version", "")}'

        if _pypi_version > version:

            msg = (
                f'{gl("New_version_available")}{_pypi_version}\r\n\r\n'
                f'{get_version_description()}\n'
                f'\n{gl("Please_update_your_current_version")}{version}\n'
            )
        else:
            msg = (
                f'{gl("No_new_version_available")}'
                f'{gl("Current_version")} {version}\r\n'
            )
    except (HTTPError, Exception) as e:
        ExceptionErrorMessage(e)

    return msg


# Get Remaining arguments
def _get_remain_arg(args: list[str] | str) -> str:
    """ Get remainig arguments values"""
    n = 0
    arg = ""
    for i in sys.argv:
        if i in args:
            arg = sys.argv[n + 1] if n + 1 < len(sys.argv) else arg
            break
        n = n + 1
    return arg


# Check version action class
class ChkVersionAction(argparse.Action):
    """Class Check version. This class call for `check_version` function"""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw,):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        p = getattr(namespace, "proxy") or _get_remain_arg(["-x", "--proxy"])
        if not p:
            config = ConfigManager()
            proxy = config.get("proxy")
        else:
            proxy = p if validate_proxy(p) else None

        print(_check_version(version("subdx-dl"), proxy))
        sys.exit(0)


# Class VideoExtractor
class VideoMetadataExtractor:
    """
    A class to extract metadata from video filenames using guessit.
    """
    @typing.no_type_check
    @staticmethod
    def extract_all(filename: str, options: str | dict[str, Any] = {}) -> dict[str, Any]:
        """
        Extract all available metadata from a video filename.

        Args:
            filename (str): The video filename to parse

        :param options:
        :type options: str|dict

        Returns:
            dict: Dictionary containing all extracted properties
        """
        all_metadata = guessit(filename, options)
        result_dict: dict[str, Any] = {}
        for key, value in all_metadata.items():
            if isinstance(value, jsonutils.Match):
                result_dict[key] = {
                    'value': value.value,
                    'raw': value.raw
                }
            else:
                result_dict[key] = value
        return result_dict

    @staticmethod
    def extract_specific(filename: str, *properties: str, options: str | dict[str, Any] = {}) -> dict[str, Any]:
        """
        Extract specific properties from a video filename.

        Args:
            filename (str): The video filename to parse
            *properties (str): Properties to extract (e.g., 'title', 'year')

        :param options:
        :type options: str|dict

        Returns:
            dict: Dictionary containing only the requested properties
        """
        all_metadata: dict[str, Any] = {}
        all_metadata = guessit(filename, options)  # type: ignore
        result_dict: dict[str, Any] = {}
        for key, value in all_metadata.items():  # type: ignore
            if isinstance(value, jsonutils.Match):
                result_dict[key] = {
                    'value': value.value,  # type: ignore
                    'raw': value.raw  # type: ignore
                }
            else:
                result_dict[key] = value

        return {prop: result_dict.get(prop) for prop in properties}

    @staticmethod
    def pretty_print(metadata: dict[str, Any]) -> None:
        """
        Pretty print the metadata dictionary.

        Args:
            metadata (dict): Metadata dictionary to print
        """

        console.print_json(data=metadata, indent=4, default=str)


# Class Config Settings
class ConfigManager:
    """
    A class to manage application configuration settings in a JSON file.

    Attributes:
        config_path (str): Path to the configuration file
        config (dict): Dictionary containing the configuration settings
    """

    def __init__(self, config_path: str = ""):
        """
        Initialize the ConfigManager with a path to the configuration file.

        Args:
            config_path (str): Path to the configuration file. Defaults to None.
        """
        self.config_path = config_path if config_path else self.get_path()
        self.config = {}

        # Load existing config if it exists
        self._load_config()

    @property
    def exists(self) -> bool:
        """ Check if exists a config file"""
        return os.path.isfile(self.config_path)

    @property
    def hasconfig(self) -> bool:
        """ Check if config is empty"""
        return bool(self.config)

    def _load_config(self) -> None:
        """Load the configuration from file or create a new one if it doesn't exist."""
        try:
            if self.exists:
                with open(self.config_path, 'r') as f:
                    self.config = json.load(f)
            else:
                self.config = {}
        except (json.JSONDecodeError, IOError) as e:
            pass
            console.print(
                f':no_entry: [bold red]{gl("Failed_to_load_configuration")}[/]{e.__class__.__name__}\n',
                emoji=True, new_line_start=True
            )
            self._save_config()
            sys.exit(1)

    def _save_config(self) -> None:
        """Save the current configuration to file."""
        if not self.exists:
            config_dir = Path(os.path.dirname(self.config_path))
            config_dir.mkdir(parents=True, exist_ok=True)
        try:
            with open(self.config_path, 'w') as f:
                json.dump(self.config, f, indent=4)
        except IOError as e:
            pass
            console.print(
                f':no_entry: [bold red]{gl("Failed_to_save_configuration")}[/]{e.__class__.__name__}\n',
                emoji=True, new_line_start=True
            )
            sys.exit(1)

    def get(self, key: str, default: Any | None = None) -> Any:
        """
        Get a configuration value by key.

        Args:
            key (str): The configuration key to retrieve
            default (Any): Default value to return if key doesn't exist

        Returns:
            The configuration value or default if key doesn't exist
        """
        return self.config.get(key, default)

    def set(self, key: str, value: Any) -> None:
        """
        Set a configuration value.

        Args:
            key (str): The configuration key to set
            value (Any): The value to set
        """
        self.config[key] = value
        self._save_config()

    def update(self, new_config: dict[str, Any]) -> None:
        """
        Update multiple configuration values at once.

        Args:
            new_config (dict): Dictionary of key-value pairs to update
        """
        self.config.update(new_config)
        self._save_config()

    def delete(self, key: str) -> None:
        """
        Delete a configuration key.

        Args:
            key (str): The configuration key to delete
        """
        if key in self.config:
            del self.config[key]
            self._save_config()

    def reset(self) -> None:
        """Reset the configuration to an empty state."""
        self.config = {}
        self._save_config()

    def get_all(self) -> dict[str, Any]:
        """
        Get all configuration settings.

        Returns:
            dict: A copy of the current configuration
        """
        return self.config.copy()

    def save_all(self, config: dict[str, Any]) -> None:
        """
        Save all configuration values.

        Args:
            dict: With all configuration values.
        """

        self.reset()
        self.config = config.copy()
        self._save_config()

    def print_config(self) -> None:
        """
        Pretty print the config dictionary.
        """
        console.print_json(data=self.config, indent=4, default=str)

    def merge_config(self, args: dict[str, Any]) -> dict[str, Any]:
        """
        Merge args values with config file

        Args:
            dict: With arguments to merge
        """
        merged: dict[str, Any] = {}
        excluded = ["browser_path", "SubX_key"]
        copy_conf = self.config.copy()
        for key in excluded:
            if key in copy_conf:
                del copy_conf[key]
        merged = {**args, **{k: v for k, v in copy_conf.items() if not args[k]}}

        return merged

    @staticmethod
    def get_path(app_name: str = "subdx-dl", file_name: str | None = "sdx-config.json") -> Path:
        """
        Get the appropriate local configuration path for the current platform.

        Args:
            app_name: Name of your application (used to create a subdirectory). Default subdx-dl
            file_name: Optional filename to append to the config path. Default sdx-config.json

        Returns:
            Path object pointing to the configuration directory or file

        Platform-specific paths:
        - Windows: %LOCALAPPDATA%\\<app_name>\\
        - macOS: ~/Library/Application Support/<app_name>/
        - Linux: ~/.config/<app_name>/
        """
        if sys.platform == "win32":
            # Windows
            base_dir = Path(f'{os.getenv("LOCALAPPDATA")}')
        elif sys.platform == "darwin":
            # macOS
            base_dir = Path.home() / "Library" / "Application Support"
        else:
            # Linux and other UNIX-like systems
            base_dir = Path.home() / ".config"

        config_dir = base_dir / app_name

        if file_name:
            return config_dir / file_name
        return config_dir


# Config action classes
class ViewConfigAction(argparse.Action):
    """Check config file class Action"""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw,):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        config = ConfigManager()
        if config.exists:
            console.print("[bold yellow]" + gl("Config_file") + "[/]", f'{config.get_path()}')
            config.print_config() if config.hasconfig else console.print(":no_entry: [bold red]" + gl("Config_is_empty") + "[/]")
        else:
            console.print(":no_entry: [bold red]" + gl("Not_exists_a_config_file") + "[/]")
        sys.exit(0)


class SaveConfigAction(argparse.Action):
    """Save allowed arguments to a config file. Existing values are update."""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw,):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        allowed_values = ["quiet", "verbose", "force", "no_choose", "no_filter", "nlines", "path", "proxy", "Season", "imdb", "lang"]
        copied_config = namespace.__dict__.copy()

        if all(not copied_config[k] for k in copied_config.keys()):
            console.print(":no_entry:[bold yellow]" + gl("Nothing_to_save") + "[/]")
            sys.exit(0)

        for k in namespace.__dict__.keys():
            if k not in allowed_values:
                del copied_config[k]

        config = ConfigManager()

        config.update(config.merge_config(copied_config)) if config.hasconfig else config.save_all(copied_config)
        if not copied_config['quiet']:
            console.print(f':heavy_check_mark: {gl("Config_was_saved")}')

        if not getattr(namespace, "search"):
            sys.exit(0)


class SetConfigAction(argparse.Action):
    """Save an option to config file"""
    @typing.no_type_check
    def __init__(self, nargs='?', **kw):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):

        if not values:
            console.print(f':no_entry: [bold red]{gl("Not_a_valid_option")}[/]', self.choices)
            sys.exit(1)

        key, value = "", None
        cf = ConfigManager()

        if values in ["quiet", "verbose", "force", "no_choose", "no_filter", "Season", "imdb", "SubX"]:
            key, value = f'{values}', bool(True)
            if values == "SubX":
                subx_key = _get_remain_arg("SubX")
                if subx_key:
                    cf.update({"SubX_key": subx_key})
                elif not cf.get("SubX_key"):
                    console.print(
                        f':warning:  {gl("Not_SubX_key")}\n'
                        f'[italic pale_turquoise4]{gl("Not_SubX_key_wiki")}[/]',
                        emoji=True, new_line_start=True
                    )
        elif values == "path":
            path = _get_remain_arg("path")
            if os.path.isdir(path) and os.access(path, os.W_OK):
                key, value = f'{values}', path
            else:
                console.print(
                    f':no_entry: [bold red]{gl("Directory")}[/][yellow]{path}[bold red] '
                    f'{gl("Directory_not_exists")}[/]'
                )
        elif values == "proxy":
            proxy = _get_remain_arg("proxy")
            if validate_proxy(proxy):
                key, value = f'{values}', proxy
            else:
                console.print(
                    f':no_entry: [bold red]{gl("Incorrect_proxy_setting").split(".")[0]}: '
                    f'[yellow]{proxy}[/]'
                )
        elif values == "nlines":
            lines = _get_remain_arg("nlines")
            key, value = f'{values}', int(lines) if lines.isnumeric() and int(lines) in range(5, 25, 5) else 10
        elif values == "lang":
            language = _get_remain_arg("lang")
            key, value = f'{values}', language if language in ["es", "en"] else "es"

        if not value:
            sys.exit(1)

        if cf.hasconfig:
            cf.set(key, value)
        else:
            cf.update({key: value})

        console.print(gl("Done"))
        sys.exit(0)


class ResetConfigAction(argparse.Action):
    """Reset an option in the config file"""
    @typing.no_type_check
    def __init__(self, nargs='?', **kw):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):

        if not values:
            console.print(":no_entry:[bold red]  " + gl("Not_a_valid_option") + "[/]", self.choices)
            sys.exit(1)

        config = ConfigManager()

        if values in ["quiet", "verbose", "force", "no_choose", "no_filter", "Season", "imdb", "path", "proxy", "nlines", "lang", "SubX"]:
            config.delete(values)

        console.print(gl("Done"))
        sys.exit(0)


# Bypasser actions
class SetBypasserConfigAction(argparse.Action):
    """Set Bypasser config"""
    @typing.no_type_check
    def __init__(self, nargs=0, **kw):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        from rich.prompt import Prompt
        cf = ConfigManager()

        browser = Prompt.ask(
            f'[bold yellow]{gl("Browser_path")}[/]\n'
            f'[italic pale_turquoise4]{gl("Browser_path_ex")}[/]',
            show_default=True,
            default=None
        )

        if browser:
            try:
                exists_browser = os.path.isfile(browser) and os.access(browser, os.X_OK)
                assert exists_browser, gl("Not_exists_browser")
                key, value = "browser_path", browser
                if cf.hasconfig:
                    cf.set(key, value)
                else:
                    cf.update({key: value})
                console.print(f'[bold yellow]{gl("Done")}[/]', new_line_start=True)
                sys.exit(0)
            except AssertionError as e:
                console.print(f':no_entry: [bold red]{e}[/]')
                sys.exit(1)
        else:
            console.print(
                f'\r\n:no_entry: [bold red]{gl("Not_browser_path")}[/]',
                emoji=True, new_line_start=False
            )
            sys.exit(1)


class BypasserAction(argparse.Action):
    """Bypasser class Action"""
    @typing.no_type_check
    def __init__(self, nargs='?', **kw,):
        super().__init__(nargs=nargs, **kw)

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        from sdx_dl.cf_bypasser.get_cf_bypass import get_cf_bypass, manual_bypasser

        cf = ConfigManager()

        if values == "manual":
            manual_bypasser()
            sys.exit(0)

        if cf.hasconfig and 'browser_path' in cf.config:
            browser = f'{cf.get("browser_path")}'
        else:
            browser = None

        if browser:
            force = True if values == "force" else False
            get_cf_bypass(browser, force)
        else:
            console.print(
                f':no_entry: [bold red]{gl("Not_browser_path")}[/]',
                emoji=True, new_line_start=True
            )
            sys.exit(1)

        sys.exit(0)


# Setting config language
config = ConfigManager()
if config.hasconfig and 'lang' in config.config:
    set_locale(config.get("lang", "es"))

# Findfiles class
extension_pattern = '(\\.[a-zA-Z0-9]+)$'
string_type = str


class InvalidPath(Exception):
    """Raised when an argument is a non-existent file or directory path
    """
    pass


class FindFiles(object):
    """Given a file, it will verify it exists. Given a folder it will descend
    one level into it and return a list of files, unless the recursive argument
    is True, in which case it finds all files contained within the path.

    The with_extension argument is a list of valid extensions, without leading
    spaces. If an empty list (or None) is supplied, no extension checking is
    performed.

    The filename_blacklist argument is a list of regexp strings to match against
    the filename (minus the extension). If a match is found, the file is skipped
    (e.g. for filtering out "sample" files). If [] or None is supplied, no
    filtering is done
    """

    def __init__(self, path: str, with_extension: list[str] | None = None, filename_blacklist: list[Any] | None = None, recursive: bool = False):

        self.path = path
        if with_extension is None:
            self.with_extension = []
        else:
            self.with_extension = with_extension
        if filename_blacklist is None:
            self.with_blacklist = []
        else:
            self.with_blacklist = filename_blacklist
        self.recursive = recursive

    @staticmethod
    def split_extension(filename: str) -> str:
        """Split extension from `filename` based in extension pattern"""
        base = re.sub(extension_pattern, "", filename)
        ext = filename.replace(base, "")
        return ext

    def findFiles(self) -> list[str]:
        """Returns list of files found at path
        """
        listfiles: list[str] = []
        if os.path.isfile(self.path):
            path = os.path.abspath(self.path)
            if self._checkExtension(path) and not self._blacklistedFilename(path):
                listfiles.append(path)
                return listfiles
            else:
                return listfiles
        elif os.path.isdir(self.path):
            return self._findFilesInPath(self.path)
        else:
            raise InvalidPath("%s is not a valid file/directory" % self.path)

    def _checkExtension(self, fname: str) -> bool:
        """Checks if the file extension is blacklisted in valid_extensions
        """
        if len(self.with_extension) == 0:
            return True

        # don't use split_extension here (otherwise valid_extensions is useless)!
        _, extension = os.path.splitext(fname)
        for cext in self.with_extension:
            cext = ".%s" % cext
            if extension == cext:
                return True
        else:
            return False

    def _blacklistedFilename(self, filepath: str) -> bool:
        """Checks if the filename (optionally excluding extension)
        matches filename_blacklist

        self.with_blacklist should be a list of strings and/or dicts:

        a string, specifying an exact filename to ignore
        "filename_blacklist": [".DS_Store", "Thumbs.db"],

        a dictionary, where each dict contains:

        Key 'match' - (if the filename matches the pattern, the filename
        is blacklisted)

        Key 'is_regex' - if True, the pattern is treated as a
        regex. If False, simple substring check is used (if
        cur['match'] in filename). Default is False

        Key 'full_path' - if True, full path is checked. If False, only
        filename is checked. Default is False.

        Key 'exclude_extension' - if True, the extension is removed
        from the file before checking. Default is False.
        """

        if len(self.with_blacklist) == 0:
            return False

        fullname = f'{os.path.split(filepath)[1]}'
        fname = self.split_extension(fullname)

        for fblacklist in self.with_blacklist:
            if isinstance(fblacklist, string_type):
                if fullname == fblacklist:
                    return True
                else:
                    continue

            if "full_path" in fblacklist and fblacklist["full_path"]:
                to_check = filepath
            else:
                if fblacklist.get("exclude_extension", False):
                    to_check = fname
                else:
                    to_check = fullname

            if fblacklist.get("is_regex", False):
                m = re.match(fblacklist["match"], to_check)
                if m is not None:
                    return True
            else:
                m = fblacklist["match"] in to_check
                if m:
                    return True
        else:
            return False

    def _findFilesInPath(self, startpath: str) -> list[str]:
        """Finds files from startpath, could be called recursively
        """
        allfiles: list[str] = []
        if not os.access(startpath, os.R_OK):
            return allfiles

        for subf in os.listdir(string_type(startpath)):
            newpath = os.path.join(startpath, subf)
            newpath = os.path.abspath(newpath)
            if os.path.isfile(newpath):
                if not self._checkExtension(subf):
                    continue
                elif self._blacklistedFilename(subf):
                    continue
                else:
                    allfiles.append(newpath)
            else:
                if self.recursive:
                    allfiles.extend(self._findFilesInPath(newpath))
        return allfiles



import os
import logging
import typing
import asyncio
import tempfile
import requests

from hashlib import md5
from rich.logging import RichHandler
from rich.traceback import install
install(show_locals=True)


def md5_hash(text: str | bytes) -> str:
    if isinstance(text, str):
        text = text.encode('utf-8')
    return md5(text).hexdigest()


@typing.no_type_check
def get_browser_init_lock() -> asyncio.Lock:
    """Get the global browser initialization lock for the current event loop."""

    # Global lock state for browser initialization
    global _global_lock_state
    _global_lock_state = {"lock": None, "loop": None}

    try:
        current_loop = asyncio.get_running_loop()
    except RuntimeError:
        return asyncio.Lock()

    if _global_lock_state["lock"] is None or _global_lock_state["loop"] != current_loop:
        _global_lock_state["lock"] = asyncio.Lock()
        _global_lock_state["loop"] = current_loop

    return _global_lock_state["lock"]


def create_logger(level: str = "DEBUG", verbose: bool = False):

    # Setting logger
    levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
    LOGGER_LEVEL = levels[4]
    LOGGER_FORMATTER_LONG = logging.Formatter('%(asctime)-12s %(levelname)-6s %(message)s', '%Y-%m-%d %H:%M:%S')
    LOGGER_FORMATTER_SHORT = logging.Formatter(fmt='%(message)s', datefmt="[%X]")

    level = level if level in levels else LOGGER_LEVEL
    temp_log_dir = tempfile.gettempdir()
    file_log = os.path.join(temp_log_dir, 'subdx-dl.log')

    global logger
    logger = logging.getLogger(__name__)
    logger.setLevel(level)

    if not verbose:
        logfile = logging.FileHandler(file_log, mode='w', encoding='utf-8')
        logfile.setFormatter(LOGGER_FORMATTER_LONG)
        logfile.setLevel(level)
        logger.addHandler(logfile)
    else:
        console = RichHandler(rich_tracebacks=True, show_path=False)
        console.setFormatter(LOGGER_FORMATTER_SHORT)
        console.setLevel(level)
        logger.addHandler(console)


create_logger()


def get_public_ip(proxy: str | None = None):
    """Get hostname public ip"""

    services = [
        'https://api.ipify.org',
        'https://checkip.amazonaws.com'
    ]
    proxies = {'http': proxy, 'https': proxy} if proxy else None

    for service in services:
        try:
            response = requests.get(service, timeout=10, proxies=proxies)
            response.raise_for_status()  # Raise an exception for bad status codes
            ip = response.text.strip()
            # Basic validation that it looks like an IP address
            if ip.count('.') == 3 and all(part.isdigit() for part in ip.split('.')):
                return ip
        except requests.RequestException as e:
            logger.debug(f"Failed to get IP from {service}: {e}")
            continue

    return None



# Copyright (C) 2024 Spheres-cu (https://github.com/Spheres-cu) subdx-dl
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

import os
import sys
import time
import shutil
import tempfile

from tempfile import NamedTemporaryFile
from urllib3.response import HTTPResponse
from requests import Response
from rarfile import Error  # type: ignore
from rarfile import RarFile, is_rarfile  # type: ignore
from zipfile import ZipFile, is_zipfile
from sdx_dl.sdxparser import args, logger
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl
from sdx_dl.sdxclasses import ConfigManager
from sdx_dl.sdxsubxapi import SubxAPI
from typing import Any
from sdx_dl.sdxutils import extract_subtitles  # type: ignore
from sdx_dl.sdxutils import (
    get_imdb_search,
    get_aadata,
    convert_date,
    get_filtered_results,
    sort_results,
    get_selected_subtitle_id,
    HTTPErrorsMessageException,
    clean_screen,
    paginate,
    Metadata,
    metadata,
    SUBDIVX_DOWNLOAD_PAGE,
    HTTPError,
    headers,
    conn
)

__all__ = ["get_subtitle_id", "get_subtitle"]


def get_subtitle_id(title: str, number: str, inf_sub: dict[str, Any], metadata: Metadata = metadata):

    """
    Get a list of subtitles of subtitles searched by ``title`` and season/episode
    ``number`` of series or movies.

    The results are ordered based on a weighing of a ``metadata`` list.

    If ``no_choose`` ``(-nc)`` is false then a list of subtitles is show for choose.

    Else the first founded subtitle `id` is choosen.

    Return the subtitle `id`
    """

    buscar = None
    res = ""
    list_Subs_Dicts: list[dict[str, Any]] = []

    if args.imdb:
        if not args.quiet:
            console.print(
                f':earth_americas: [bold yellow]{gl("Searching_IMDB")} {title} {number}',
                new_line_start=True, emoji=True
            )
        logger.debug(f'Searching in IMDB: {title} {number}')
        search = get_imdb_search(title, number, inf_sub)
        buscar = search
        if buscar is not None and inf_sub['type'] == 'episode':
            title = buscar.replace(number, "").strip()
        logger.debug(f'IMDB Search result: {buscar}')

        if not args.quiet:
            clean_screen()
            imdb_search = buscar if buscar is not None else "Ninguno"
            console.print(
                f':information_source:  [bold yellow]{gl("Search_terms_from_IMDB")}[/]{imdb_search}',
                new_line_start=True, emoji=True)
            time.sleep(0.5)

    if buscar is None:
        buscar = f"{title} {number}".strip()

    if not args.quiet:
        console.print("\r")
    logger.debug(f'Searching subtitles for: {title} {number.upper()}')

    with console.status(f'{gl("Searching_subtitles_for")}{title} {number.upper()}') as status:
        status.start() if not args.quiet else status.stop()
        if args.SubX:
            cf = ConfigManager()
            if cf.hasconfig and "SubX_key" in cf.config:
                sbx = SubxAPI(cf.get("SubX_key", default=""))
                sbx.query(buscar)
                json_aaData = sbx.from_subx_aadata()
            else:
                console.print(
                    f':no_entry: {gl("Not_SubX_key")}\r\n'
                    f'[italic pale_turquoise4]{gl("Not_SubX_key_wiki")}[/]',
                    emoji=True, new_line_start=False
                )
                sys.exit(1)
        else:
            json_aaData = get_aadata(buscar)

    if json_aaData["iTotalRecords"] == 0:
        if not args.quiet:
            console.print(f':no_entry: [bold red]{gl("Not_subtitles_records_found_for")}[/] [yellow]{buscar}[/]')
        logger.debug(f'Not subtitles records found for: {buscar}')
        return res
    else:
        logger.debug(f'Found subtitles records for: {buscar}')

    # Get Json Data Items
    aaData_Items = json_aaData['aaData']

    if aaData_Items is not None:
        list_Subs_Dicts = convert_date(aaData_Items)
    else:
        if not args.quiet:
            console.print(f':no_entry: [bold red]{gl("No_suitable_data_were_found_for")} [yellow]{buscar}[/]')
        logger.debug(f'No suitable data were found for: "{buscar}"')
        return res

    # only include results for this specific serie / episode
    # ie. search terms are in the title of the result item

    if args.imdb or args.no_filter:
        filtered_list_Subs_Dicts = list_Subs_Dicts
    else:
        filtered_list_Subs_Dicts = get_filtered_results(title, number, inf_sub, list_Subs_Dicts)

    if not filtered_list_Subs_Dicts:
        if not args.quiet:
            console.print(f':no_entry: [bold red]{gl("No_suitable_data_were_found_for")} [yellow]{buscar}[/]')
        logger.debug(f'No suitable data were found for: "{buscar}"')
        return res

    if metadata.hasdata:
        results = sort_results(filtered_list_Subs_Dicts, metadata)
    else:
        results = sorted(filtered_list_Subs_Dicts, key=lambda item: (item['descargas']), reverse=True)

    # Print subtitles search infos
    # Construct Table for console output

    table_title = f'{title} {number.upper()}'
    results_pages = paginate(results, 10)

    if (not args.no_choose):
        res = get_selected_subtitle_id(table_title, results, metadata)
        return res
    else:
        # get first subtitle
        res = f"{results_pages['pages'][0][0]['id']}"

    return res


def get_subtitle(subid: str, topath: str):
    """Download a subtitle with id ``subid`` to a destination ``path``."""

    url = f"{SUBDIVX_DOWNLOAD_PAGE + 'descargar.php?id=' + subid}"
    subx_url = f"https://subx-api.duckdns.org/api/subtitles/{subid}/download"

    if not args.quiet:
        clean_screen()

    temp_file = NamedTemporaryFile(delete=False)
    download_url = None
    data: bytes = b''

    # get direct download link
    if not args.quiet:
        console.print(gl("Downloading_Subtitle"), emoji=True, new_line_start=True)
    logger.debug(f"Trying Download from link: {url if not args.SubX else subx_url}")

    if args.SubX:
        cf = ConfigManager()
        if cf.hasconfig and "SubX_key" in cf.config:
            sbx = SubxAPI(cf.get("SubX_key", default=""))
            download_url = sbx.get(subid)
    else:
        try:
            download_url = conn.request('GET', url, headers=headers)
        except HTTPError as e:
            HTTPErrorsMessageException(e)
            sys.exit(1)

    if download_url:
        if not args.SubX and isinstance(download_url, HTTPResponse):
            logger.debug(f"Downloaded from: {SUBDIVX_DOWNLOAD_PAGE}{download_url.geturl()}")
            data = download_url.data
        elif isinstance(download_url, Response):
            logger.debug(f"Downloaded from: SubX: {subx_url}")
            data = download_url.content

        temp_file.write(data)
        temp_file.seek(0)
        # Checking if the file is zip or rar then decompress
        try:
            if is_zipfile(temp_file):
                compressed_sub_file = ZipFile(temp_file)
                extract_subtitles(compressed_sub_file, topath)
            elif is_rarfile(temp_file):
                compressed_sub_file = RarFile(temp_file)
                extract_subtitles(compressed_sub_file, topath)
        except (Error):
            console.clear()
            temp_dir = tempfile.gettempdir()
            shutil.copyfile(os.path.join(temp_dir, temp_file.name), os.path.join(topath, f'{subid}.rar'))

            console.print(
                f':warning:  [bold red]{gl("Cannot_find_a_working_tool")}[bold yellow]{gl("Install_rar")}[/]',
                emoji=True, new_line_start=True
            )
            logger.debug("Cannot find a working tool, please install rar decompressor tool")
            logger.debug(f"File downloaded to: {os.path.join(topath, f'{subid}.rar')}")
    else:
        temp_file.close()
        os.unlink(temp_file.name)
        logger.error(f'No suitable subtitle download for : "{url}"')
        if not args.quiet:
            console.print(
                f':cross_mark:  [bold red]{gl("No_suitable_subtitle_to_download")}[/]',
                emoji=True, new_line_start=True
            )
        sys.exit(1)
        time.sleep(2)

    # Cleaning
    temp_file.close()
    os.unlink(temp_file.name)



# type:ignore

import os
import re
import sys
import time
import tempfile

from pathlib import Path
from sdx_dl.cf_bypasser.CloudflareBypasser import CloudflareBypasser
from DrissionPage import ChromiumPage, ChromiumOptions
from sdx_dl.cf_bypasser.cache.cookie_cache import CookieCache
from sdx_dl.cf_bypasser.utils.misc import md5_hash, logger, get_public_ip
from sdx_dl.sdxconsole import console
from sdx_dl.sdxlocale import gl
from urllib.parse import urlparse
from typing import Any

SUBDIVX_PAGE = 'https://www.subdivx.com/'


def get_cache_path(app_name: str = "subdx-dl", file_name: str | None = "sdx_cache_connection.json") -> Path:
    """Get platform-specific cache directory"""
    if sys.platform == "win32":
        base_dir = Path(f'{os.getenv("LOCALAPPDATA")}')
    elif sys.platform == "darwin":
        base_dir = Path.home() / "Library" / "Caches"
    else:
        base_dir = Path.home() / ".cache"

    cache_dir = base_dir / app_name

    if not os.path.isdir(cache_dir):
        cache_dir.mkdir(parents=True, exist_ok=True)

    if file_name:
        return cache_dir / file_name
    return cache_dir


def get_chromium_options(browser_path: str, arguments: list[str]) -> ChromiumOptions:
    """
    Configures and returns Chromium options.

    :param browser_path: Path to the Chromium browser executable.
    :param arguments: List of arguments for the Chromium browser.
    :return: Configured ChromiumOptions instance.
    """
    options = ChromiumOptions().auto_port()
    options.set_paths(browser_path=browser_path)
    for argument in arguments:
        options.set_argument(argument)
    return options


def get_cf_bypass(browser: str = "", force: bool = False):
    """Try get cf credentials"""

    browser_path = browser
    sdx_cache_path = get_cache_path()

    # Arguments to make the browser better for automation and less detectable.
    arguments = [
        "-no-first-run",
        "-force-color-profile=srgb",
        "-metrics-recording-only",
        "-password-store=basic",
        "-use-mock-keychain",
        "-export-tagged-pdf",
        "-no-default-browser-check",
        "-disable-background-mode",
        "-enable-features=NetworkService,NetworkServiceInProcess,LoadCryptoTokenExtension,PermuteTLSExtensions",
        "-disable-features=FlashDeprecationWarning,EnablePasswordsAccountStorage",
        "-deny-permission-prompts",
        "-disable-gpu",
        "-accept-lang=en-US"
    ]

    options = get_chromium_options(browser_path, arguments)

    # Try to get cached cookies first
    cookie_cache = CookieCache(cache_file=f'{sdx_cache_path}')
    cookie_cache.clear_expired()
    local_address = get_public_ip()
    hostname = urlparse(SUBDIVX_PAGE).netloc
    cache_key = md5_hash(hostname + local_address if local_address else "")
    cached = cookie_cache.get(cache_key)

    if not cached or cached.is_expired() or force:
        logger.debug(f"No cached cookies, expired or force bypasser for {cache_key}, generating new ones...")

        # Initialize the browser
        console.print(
            f':robot: [italic yellow]{gl("Getting_connection")}[/]',
            emoji=True, new_line_start=True
        )
        time.sleep(5)
        driver = ChromiumPage(addr_or_opts=options)
        try:
            logger.debug(f'Navigating to the {SUBDIVX_PAGE} page.')
            driver.get(SUBDIVX_PAGE)

            # Where the bypass starts
            logger.debug('Starting Cloudflare bypass.')
            cf_bypasser = CloudflareBypasser(driver)

            # If you are solving an in-page captcha (like the one here: https://seleniumbase.io/apps/turnstile), use cf_bypasser.click_verification_button() directly instead of cf_bypasser.bypass().
            # It will automatically locate the button and click it. Do your own check if needed.

            cf_bypasser.bypass()
            if not cf_bypasser.is_bypassed():
                logger.error("Bypass failed.")
                driver.quit()
                sys.exit(1)
            else:
                console.print(":heavy_check_mark:  Bypass successful!", emoji=True)

            cookies = driver.cookies(as_dict=True, all_info=True)
            sdx_dc_path = os.path.join(tempfile.gettempdir(), 'sdx_data_connection.json')

            data_cache = {
                "user_agent": driver.user_agent,
                "cookies": cookies
            }

            logger.debug("Saving credentials...")
            try:
                cookie_cache.set(cache_key, data_cache["cookies"], data_cache["user_agent"], ttl_hours=24)
                with open(sdx_dc_path, 'w') as file:
                    file.write('')
                    file.close()
            except Exception as e:
                logger.error(f"Failed to save credential file: {e}")

            logger.debug("Saved cookies successfully")

            # Sleep for a while to let the user see the result if needed
            time.sleep(5)
        except Exception as e:
            logger.error("An error occurred: %s", str(e))
        finally:
            logger.debug('Closing the browser.')
            driver.quit()
    else:
        logger.debug(f"Using Cached cookies for {cache_key}")
        console.print(
            f':white_check_mark: {gl("Still_cached_cookies")}\n'
            f'[bold yellow]{gl("Expires_at")}[/]{cached.expires_at.strftime("%H:%M:%S %d/%m/%Y")}',
            emoji=True
        )


def manual_bypasser():
    from rich.prompt import Prompt

    try:
        cf_clearance = Prompt.ask(
            f'[bold yellow]{gl("cf_clearance_cookie")}[/]',
            show_default=False,
            default=""
        )
        if cf_clearance:
            assert re.match(r'[a-zA-Z0-9\.\-\_]+$', cf_clearance), gl("Invalid_cf_clearance")
        else:
            raise AssertionError(gl("Invalid_cf_clearance"))

        sdx_cookie = Prompt.ask(
            f'[bold yellow]{gl("sdx_cookie")}[/]',
            show_default=False,
            default=""
        )
        if sdx_cookie:
            assert re.match(r'^[a-zA-Z0-9]+$', sdx_cookie), gl("Invalid_sdx")
        else:
            raise AssertionError(gl("Invalid_sdx"))

        user_agent = Prompt.ask(
            f'[bold yellow]{gl("User_agent")}[/]',
            show_default=False,
            default=""
        )
        if user_agent:
            browser_patterns = [
                r'Mozilla/\d+\.\d+',
                r'AppleWebKit/\d+\.\d+',
                r'Chrome/\d+',
                r'Safari/\d+',
                r'Firefox/\d+',
                r'Edge/\d+'
            ]
            valid_user_agent = any(re.search(pattern, user_agent) for pattern in browser_patterns)
            assert valid_user_agent, gl("Invalid_user_agent")
        else:
            raise AssertionError(gl("Invalid_user_agent"))

        try:
            sdx_cache_path = get_cache_path()
            cookie_cache = CookieCache(cache_file=f'{sdx_cache_path}')
            cookie_cache.clear_expired()
            local_address = get_public_ip()
            hostname = urlparse(SUBDIVX_PAGE).netloc
            cache_key = md5_hash(hostname + local_address if local_address else "")
            cookies = {"cf_clearance": cf_clearance, "sdx": sdx_cookie}
            data_cache: dict[str, Any] = {
                "user_agent": user_agent,
                "cookies": cookies
            }

            cookie_cache.set(cache_key, data_cache["cookies"], data_cache["user_agent"], ttl_hours=24)
            console.print(f'[bold green]{gl("Done")}[/]', new_line_start=True)
        except Exception as e:
            console.print(
                f':no_entry: [bold red]{gl("Failed_to_save_cache")}[/]\n'
                f'Error: {e}',
                emoji=True, new_line_start=False
            )
            sys.exit(1)
    except AssertionError as e:
        console.print(f':no_entry: [bold red]{e}[/]')
        sys.exit(1)
